--[[
   cellmap.lua

   Copyright 2017 wesley werner <wesley.werner@gmail.com>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

References:

  Excellent noise generation page:
  http://lodev.org/cgtutor/randomnoise.html

  Nice tutorial implementing cellular automata:
  https://gamedevelopment.tutsplus.com/tutorials/generate-random-cave-levels-using-cellular-automata--gamedev-9664

  Good old rogue basin for cave generation:  http://www.roguebasin.com/index.php?title=Cellular_Automata_Method_for_Generating_Random_Cave-Like_Levels

]]--

local map = {}

--- Returns an empty 2D array of width and height.
function map:array(width, height)

  local a = {}

  for x=1, width do
    a[x] = {}
    for y=1, height do
      a[x][y] = false
    end
  end

  return a

end

-- Fills an array with noise.
function map:noise(a, seed)

  seed = seed or os.time()
  math.randomseed(seed)

  local width = #a
  local height = #a[1]

  for x=1, width do
    for y=1, height do
      a[x][y] = math.random()
    end
  end

end

--- Populate an array with random bits weighed towards density (0..1)
function map:populate(a, seed, density)

  seed = seed or os.time()
  math.randomseed(seed)

  local width = #a
  local height = #a[1]

  for x=1, width do
    for y=1, height do
      if math.random() < density then
        a[x][y] = true
      else
        a[x][y] = false
      end
    end
  end

end

-- Place random blocks in the array
function map:populateBlocks(a, seed, density)

  seed = seed or os.time()
  math.randomseed(seed)

  local width = #a
  local height = #a[1]
  local iterations = width * density

  for iteration=1, iterations do
  
    -- random position
    local blockx = math.random(1, width)
    local blocky = math.random(1, height)
    local blockw = math.floor(math.random(1, width * 0.2))  -- % of size
    local blockh = math.floor(math.random(1, height * 0.2))
    
    for offsetx=1, blockw do
      for offsety=1, blockh do

        -- wrap around
        local x = math.max(1, (blockx + offsetx) % width)
        local y = math.max(1, (blocky + offsety) % height)
        a[x][y] = true
      
      end
    end
  end

end

--- Count living neighbouring cells in an array.
function map:countNeighbours(a, px, py)

  local width = #a
  local height = #a[1]

  local count = 0
  for ix=-1,1 do
    for iy=-1,1 do
      if ix == 0 and iy == 0 then
        -- skip counting ourselves
      else
        local nx = px+ix
        local ny = py+iy
        -- consider a boundary as a positive.
        -- this helps to fill in the edges of the map.
        if nx < 1 or ny < 1 or nx > width or ny > height then
          -- comment the line below to generate open-sea like islands
          count = count + 1
        elseif a[nx][ny] then
          -- count if this neighbour is positive
          count = count + 1
        end
      end
    end
  end
  return count

end

--- Apply cellular automata to a map
function map:cellulate(a, iterations)

  -- tweakers:
  -- the deathLimit is a good tweak.

  -- minimum living neighbours to survive (for living cells)
  local deathLimit = 4
  -- minimum living neighbours to reproduce (for dead cells)
  local birthLimit = 4

  local width = #a
  local height = #a[1]

  -- work-on copy
  local b = self:array(width, height)

  for iter=1, iterations do
    for x=1, width do
      for y=1, height do
        -- this cell is alive if truthy
        local alive = a[x][y]
        -- copy the cell
        b[x][y] = alive
        -- count neighbours at this point
        local neighbours = self:countNeighbours(a, x, y)
        -- alone cells die off without enough neighbours
        if alive and neighbours < deathLimit then
          -- the cell dies
          b[x][y] = false
        elseif not alive and neighbours > birthLimit then
          -- a new cell is born
          b[x][y] = true
        end
      end
    end

    -- copy results
    for x=1, width do
      for y=1, height do
        a[x][y] = b[x][y]
      end
    end

  end -- do another iteration

end

--- Return a new generated map
function map:generate(width, height, seed, density, iterations)

  local data = {}
  data.width = width
  data.height = height
  data.seed = seed
  data.density = density
  data.iterations = iterations
  
  -- these generate the contour map
  data.contour = self:array(width, height)
  self:populate(data.contour, seed, density)
  -- add more variety to the countour by placing some random blocks
  self:populateBlocks(data.contour, seed, density)
  self:cellulate(data.contour, iterations)
  
  -- these generate the depth map:
  -- it is a contour map with a higher density
  data.depth = self:array(width, height)
  self:populate(data.depth, seed, density + 0.1)
  self:populateBlocks(data.depth, seed, density + 0.1)
  self:cellulate(data.depth, iterations)
  return data

end

return map
